from contextlib import suppress

from .api.async_wrapper import PostalPyAPI as AsyncPostalPyAPI
from .api.wrapper import PostalPyAPI
from .smtp.wrapper import PostalPySMTP

with suppress(ImportError):
    from .smtp.async_wrapper import PostalPySMTP as AsyncPostalPySMTP



import logging
from uuid import uuid4

try:
    from aiosmtplib import (SMTP,
                            SMTPResponse)
except ImportError:
    SMTP = None
    SMTPResponse = None

from .base import PostalPySMTPBase
from .schemas import SMTPMessageSchema


class PostalPySMTP(PostalPySMTPBase):
    def __init__(self, hostname: str, username: str, password: str, port: int = 25, use_tls: bool = True,
                 timeout: int = 5, level: logging = logging.INFO):
        """
        Asynchronous SMTP client for sending messages via a configured SMTP relay.
        Requires `aiosmtplib`. Install with `pip install postal_py[smtp]`.
        """
        if SMTP is None:
            raise ImportError(
                'Async SMTP support is not available. To enable it, install extra dependencies with:\n'
                '    pip install postal_py[smtp]'
            )
        super().__init__(hostname=hostname, port=port, username=username, password=password, use_tls=use_tls,
                         timeout=timeout, level=level)

    async def send_message(self, data: SMTPMessageSchema) -> tuple[dict[str, SMTPResponse], str]:
        request_id = uuid4().hex
        self._logger.info('Request=%s data=%s', request_id, data.model_dump(
            exclude_none=True, exclude={'plain_body', 'html_body', 'attachments'}
        ))
        message = self._prepare_message(data=data)
        smtp = SMTP(hostname=self._hostname, port=self._port, timeout=self._timeout, start_tls=self._use_tls)
        await smtp.connect()
        await smtp.login(self._username, self._password)
        result = await smtp.sendmail(data.from_, data.to + data.cc + data.bcc, message.as_string())
        await smtp.quit()
        self._logger.info('Response=%s result=%s', request_id, result)
        return result



from pydantic import (BaseModel as PydanticBaseModel,
                      Field)


class BaseModel(PydanticBaseModel):
    model_config = {
        'populate_by_name': True
    }


class SMTPAttachmentSchema(BaseModel):
    filename: str
    content_type: str = 'application/octet-stream'
    data: str | bytes


class SMTPMessageSchema(BaseModel):
    to: list[str] = Field(default_factory=list)
    cc: list[str] = Field(default_factory=list)
    bcc: list[str] = Field(default_factory=list)
    from_: str = Field(None, alias='from')
    subject: str | None = None
    reply_to: str | None = None
    plain_body: str | None = None
    html_body: str | None = None
    attachments: list[SMTPAttachmentSchema] = Field(default_factory=list)
    headers: dict[str, str] = Field(default_factory=dict)



class PostalPyAPIError(Exception):
    """Base exception class for all PostalPy API-related errors."""


class PostalPyInvalidServerAPIKeyError(PostalPyAPIError):
    """The API token provided in X-Server-API-Key was not valid."""


class PostalPyAccessDeniedError(PostalPyAPIError):
    """Must be authenticated as a server"""


class PostalPyMessageNotFoundError(PostalPyAPIError):
    """No message found matching provided ID"""


class PostalPyValidationError(PostalPyAPIError):
    """The provided data was not sufficient to send an email"""


class PostalPyNoRecipientsError(PostalPyAPIError):
    """There are no recipients defined to received this message"""


class PostalPyNoContentError(PostalPyAPIError):
    """There is no content defined for this e-mail"""


class PostalPyTooManyToAddressesError(PostalPyAPIError):
    """The maximum number of To addresses has been reached (maximum 50)"""


class PostalPyTooManyCCAddressesError(PostalPyAPIError):
    """The maximum number of CC addresses has been reached (maximum 50)"""


class PostalPyTooManyBCCAddressesError(PostalPyAPIError):
    """The maximum number of BCC addresses has been reached (maximum 50)"""


class PostalPyFromAddressMissingError(PostalPyAPIError):
    """The From address is missing and is required"""


class PostalPyUnauthenticatedFromAddressError(PostalPyAPIError):
    """The From address is not authorised to send mail from this server"""


class PostalPyAttachmentMissingNameError(PostalPyAPIError):
    """An attachment is missing a name"""


class PostalPyAttachmentMissingDataError(PostalPyAPIError):
    """An attachment is missing data"""


class PostalPyConnectTimeoutError(PostalPyAPIError):
    """The request to the Postal API timed out while attempting to connect."""


class PostalPyMovedPermanentlyError(PostalPyAPIError):
    """The resource has been permanently moved (HTTP 301). Try using HTTPS instead of HTTP."""


class PostalPyPermanentRedirectError(PostalPyAPIError):
    """The resource has been permanently redirected (HTTP 308). Try using HTTPS instead of HTTP."""


class PostalPyInternalServerError(PostalPyAPIError):
    """The Postal server encountered an internal error (HTTP 500)."""


class PostalPyServiceUnavailableError(PostalPyAPIError):
    """The Postal service is temporarily unavailable (HTTP 503)."""


class PostalPyUnknownError(PostalPyAPIError):
    """An unknown error occurred while communicating with the Postal API."""



import logging

from niquests.models import Response

from .exceptions import (PostalPyAccessDeniedError,
                         PostalPyAttachmentMissingDataError,
                         PostalPyAttachmentMissingNameError,
                         PostalPyFromAddressMissingError,
                         PostalPyInternalServerError,
                         PostalPyInvalidServerAPIKeyError,
                         PostalPyMessageNotFoundError,
                         PostalPyMovedPermanentlyError,
                         PostalPyNoContentError,
                         PostalPyNoRecipientsError,
                         PostalPyPermanentRedirectError,
                         PostalPyServiceUnavailableError,
                         PostalPyTooManyBCCAddressesError,
                         PostalPyTooManyCCAddressesError,
                         PostalPyTooManyToAddressesError,
                         PostalPyUnauthenticatedFromAddressError,
                         PostalPyUnknownError,
                         PostalPyValidationError)
from .schemas import (ResponseCode,
                      ResponseSchema,
                      ResponseStatus)


class PostalPyAPIBase:
    def __init__(self, base_url: str, api_key: str, timeout: int, level: logging):
        self._base_url = base_url
        self._timeout = timeout
        self._headers = {
            'X-Server-API-Key': api_key,
            'Content-Type': 'application/json'
        }
        self._logger = logging.getLogger('PostalPyAPI')
        self._logger.setLevel(level)

    def _handle_response(self, response: Response, request_id: str) -> ResponseSchema:
        if response.status_code != 200:
            exception = {
                301: PostalPyMovedPermanentlyError,
                308: PostalPyPermanentRedirectError,
                500: PostalPyInternalServerError,
                503: PostalPyServiceUnavailableError
            }.get(response.status_code, PostalPyUnknownError)
            self._logger.error('Response=%s status_code=%s reason=%s: %s', request_id, response.status_code,
                               response.reason or 'No reason', exception.__doc__)
            raise exception(exception.__doc__)
        json_response = response.json()
        result = ResponseSchema(**json_response)
        if result.status in (ResponseStatus.error, ResponseStatus.parameter_error):
            exception = {
                ResponseCode.ACCESS_DENIED: PostalPyAccessDeniedError,
                ResponseCode.INVALID_SERVER_API_KEY: PostalPyInvalidServerAPIKeyError,
                ResponseCode.MESSAGE_NOT_FOUND: PostalPyMessageNotFoundError,
                ResponseCode.VALIDATION_ERROR: PostalPyValidationError,
                ResponseCode.NO_RECIPIENTS: PostalPyNoRecipientsError,
                ResponseCode.NO_CONTENT: PostalPyNoContentError,
                ResponseCode.TOO_MANY_TO_ADDRESSES: PostalPyTooManyToAddressesError,
                ResponseCode.TOO_MANY_CC_ADDRESSES: PostalPyTooManyCCAddressesError,
                ResponseCode.TOO_MANY_BCC_ADDRESSES: PostalPyTooManyBCCAddressesError,
                ResponseCode.FROM_ADDRESS_MISSING: PostalPyFromAddressMissingError,
                ResponseCode.UNAUTHENTICATED_FROM_ADDRESS: PostalPyUnauthenticatedFromAddressError,
                ResponseCode.ATTACHMENT_MISSING_NAME: PostalPyAttachmentMissingNameError,
                ResponseCode.ATTACHMENT_MISSING_DATA: PostalPyAttachmentMissingDataError
            }[result.data.code]
            self._logger.error('Response=%s status_code=%s json_response=%s',
                               request_id, response.status_code, json_response)
            raise exception(json_response)
        self._logger.info('Response=%s status_code=%s json_response=%s',
                          request_id, response.status_code, json_response)
        return result



import logging
from typing import Any
from uuid import uuid4

from niquests import Session
from niquests.exceptions import ConnectTimeout

from .base import PostalPyAPIBase
from .exceptions import PostalPyConnectTimeoutError
from .schemas import (MessageExpansion,
                      RequestMessageDetailsSchema,
                      RequestMessageSchema,
                      RequestRawMessageSchema,
                      ResponseSchema)


class PostalPyAPI(PostalPyAPIBase):
    def __init__(self, base_url: str, api_key: str, timeout: int = 5, level: logging = logging.INFO):
        super().__init__(base_url=base_url, api_key=api_key, timeout=timeout, level=level)
        self._session = Session(base_url=self._base_url, timeout=self._timeout)
        self._session.headers = self._headers

    def _send_request(self, url: str, json: dict[str, Any]) -> ResponseSchema:
        request_id = uuid4().hex
        self._logger.info('Request=%s url=%s json=%s', request_id, url, json)
        try:
            response = self._session.post(url=url, json=json)
        except ConnectTimeout as e:
            self._logger.exception('Response=%s: %s', request_id, PostalPyConnectTimeoutError.__doc__)
            raise PostalPyConnectTimeoutError(e)
        return self._handle_response(response=response, request_id=request_id)

    def get_message_details(self, data: RequestMessageDetailsSchema) -> ResponseSchema:
        """
        https://apiv1.postalserver.io/controllers/messages/message.html
        """
        url = '/api/v1/messages/message'
        expansions = [e for e in data.expansions or ()]
        if MessageExpansion.all in data.expansions:
            expansions = True
        json = {
            'id': data.id,
            '_expansions': expansions
        }
        return self._send_request(url=url, json=json)

    def get_message_deliveries(self, id: int) -> ResponseSchema:
        """
        https://apiv1.postalserver.io/controllers/messages/deliveries.html
        """
        url = '/api/v1/messages/deliveries'
        json = {'id': id}
        return self._send_request(url=url, json=json)

    def send_message(self, data: RequestMessageSchema) -> ResponseSchema:
        """
        https://apiv1.postalserver.io/controllers/send/message.html
        """
        url = '/api/v1/send/message'
        json = data.model_dump(exclude_none=True, by_alias=True)
        return self._send_request(url=url, json=json)

    def send_raw_message(self, data: RequestRawMessageSchema) -> ResponseSchema:
        """
        https://apiv1.postalserver.io/controllers/send/raw.html
        """
        url = '/api/v1/send/raw'
        json = data.model_dump(exclude_none=True)
        return self._send_request(url=url, json=json)

    def close(self):
        self._session.close()



import base64
import logging
from email.message import EmailMessage

from .schemas import SMTPMessageSchema


class PostalPySMTPBase:
    def __init__(self, hostname: str, port: int, username: str, password: str, use_tls: bool, timeout: int,
                 level: logging):
        self._hostname = hostname
        self._port = port
        self._username = username
        self._password = password
        self._use_tls = use_tls
        self._timeout = timeout
        self._logger = logging.getLogger('PostalPySMTP')
        self._logger.setLevel(level)

    @staticmethod
    def _prepare_message(data: SMTPMessageSchema) -> EmailMessage:
        message = EmailMessage()
        message['From'] = data.from_
        if data.to:
            message['To'] = ', '.join(data.to)
        if data.cc:
            message['Cc'] = ', '.join(data.cc)
        if data.subject is not None:
            message['Subject'] = data.subject
        if data.reply_to is not None:
            message['Reply-To'] = data.reply_to
        if data.plain_body is not None:
            message.set_content(data.plain_body)
        if data.html_body is not None:
            message.add_alternative(data.html_body, subtype='html')
        for attachment in data.attachments:
            maintype, subtype = attachment.content_type.split('/')
            if isinstance(attachment.data, bytes):
                message.add_attachment(attachment.data, maintype=maintype, subtype=subtype,
                                       filename=attachment.filename)
            else:
                message.add_attachment(base64.b64decode(attachment.data), maintype=maintype, subtype=subtype,
                                       filename=attachment.filename)
        for k, v in data.headers.items():
            message[k] = v
        return message



import logging
import ssl
from smtplib import SMTP
from typing import Any
from uuid import uuid4

from .base import PostalPySMTPBase
from .schemas import SMTPMessageSchema


class PostalPySMTP(PostalPySMTPBase):
    def __init__(self, hostname: str, username: str, password: str, port: int = 25, use_tls: bool = True,
                 timeout: int = 5, level: logging = logging.INFO):
        super().__init__(hostname=hostname, port=port, username=username, password=password, use_tls=use_tls,
                         timeout=timeout, level=level)

    def send_message(self, data: SMTPMessageSchema) -> dict[str, Any]:
        request_id = uuid4().hex
        self._logger.info('Request=%s data=%s', request_id, data.model_dump(
            exclude_none=True, exclude={'plain_body', 'html_body', 'attachments'}
        ))
        message = self._prepare_message(data=data)
        with SMTP(host=self._hostname, port=self._port, timeout=self._timeout) as smtp:
            if self._use_tls:
                context = ssl.create_default_context()
                smtp.starttls(context=context)
            smtp.login(user=self._username, password=self._password)
            result = smtp.send_message(msg=message, from_addr=data.from_, to_addrs=data.to + data.cc + data.bcc)
            self._logger.info('Response=%s result=%s', request_id, result)
            return result



from enum import Enum
from typing import Any

from pydantic import (BaseModel as PydanticBaseModel,
                      ConfigDict,
                      Field,
                      conlist)


class BaseModel(PydanticBaseModel):
    model_config = {
        'populate_by_name': True
    }


class ResponseStatus(str, Enum):
    success = 'success'
    parameter_error = 'parameter-error'
    error = 'error'


class ResponseCode(str, Enum):
    ACCESS_DENIED = 'AccessDenied'
    INVALID_SERVER_API_KEY = 'InvalidServerAPIKey'
    MESSAGE_NOT_FOUND = 'MessageNotFound'
    VALIDATION_ERROR = 'ValidationError'
    NO_RECIPIENTS = 'NoRecipients'
    NO_CONTENT = 'NoContent'
    TOO_MANY_TO_ADDRESSES = 'TooManyToAddresses'
    TOO_MANY_CC_ADDRESSES = 'TooManyCCAddresses'
    TOO_MANY_BCC_ADDRESSES = 'TooManyBCCAddresses'
    FROM_ADDRESS_MISSING = 'FromAddressMissing'
    UNAUTHENTICATED_FROM_ADDRESS = 'UnauthenticatedFromAddress'
    ATTACHMENT_MISSING_NAME = 'AttachmentMissingName'
    ATTACHMENT_MISSING_DATA = 'AttachmentMissingData'


class MessageExpansion(str, Enum):
    all = 'all'
    status = 'status'
    details = 'details'
    inspection = 'inspection'
    plain_body = 'plain_body'
    html_body = 'html_body'
    attachments = 'attachments'
    headers = 'headers'
    raw_message = 'raw_message'
    activity_entries = 'activity_entries'


class RequestAttachmentSchema(BaseModel):
    name: str
    content_type: str | None = None
    data: str


class ResponseAttachmentSchema(BaseModel):
    filename: str
    content_type: str | None = None
    data: str
    size: int
    hash: str


class ResponseStatusSchema(BaseModel):
    status: str
    last_delivery_attempt: float
    held: bool
    hold_expiry: Any


class ResponseDetailsSchema(BaseModel):
    rcpt_to: str
    mail_from: str
    subject: str | None
    message_id: str
    timestamp: float
    direction: str
    size: int
    bounce: bool
    bounce_for_id: int
    tag: str | None
    received_with_ssl: bool


class ResponseInspectionSchema(BaseModel):
    inspected: bool
    spam: bool
    spam_score: float
    threat: bool
    threat_details: Any | None


class ResponseHeadersSchema(BaseModel):
    model_config = ConfigDict(extra='allow')

    received: list[str]
    date: list[str]
    from_: list[str] = Field(..., alias='from')
    to: list[str]
    message_id: list[str] = Field(..., alias='message-id')
    subject: list[str] | None = None
    mime_version: list[str] = Field(..., alias='mime-version')
    content_type: list[str] = Field(..., alias='content-type')
    content_transfer_encoding: list[str] = Field(..., alias='content-transfer-encoding')


class ResponseLoadSchema(BaseModel):
    ip_address: str
    user_agent: str
    timestamp: str


class ResponseClickSchema(ResponseLoadSchema):
    url: str


class ResponseActivitySchema(BaseModel):
    loads: list[ResponseLoadSchema]
    clicks: list[ResponseClickSchema]


class ResponseMessageDataSchema(BaseModel):
    id: int | None = None
    token: str | None = None
    status: ResponseStatusSchema | None = None
    details: ResponseDetailsSchema | None = None
    inspection: ResponseInspectionSchema | None = None
    plain_body: str | None = None
    html_body: str | None = None
    attachments: list[ResponseAttachmentSchema] | None = None
    headers: ResponseHeadersSchema | None = None
    raw_message: str | None = None
    activity_entries: ResponseActivitySchema | None = None
    message: str | None = None
    code: str | None = None


class ResponseMessageEntrySchema(BaseModel):
    id: int
    token: str


class ResponseMessagesDataSchema(BaseModel):
    message_id: str
    messages: dict[str, ResponseMessageEntrySchema]


class ResponseStructureDataSchema(BaseModel):
    id: int
    status: str
    details: str
    output: str
    sent_with_ssl: bool
    log_id: str
    time: float
    timestamp: float


class ResponseSchema(BaseModel):
    status: ResponseStatus
    time: float
    flags: dict[str, Any]
    data: ResponseMessageDataSchema | list[ResponseStructureDataSchema] | ResponseMessagesDataSchema


class RequestMessageDetailsSchema(BaseModel):
    id: int
    expansions: set[MessageExpansion] | None = None


class RequestMessageSchema(BaseModel):
    to: conlist(str, max_length=50) | None = None
    cc: conlist(str, max_length=50) | None = None
    bcc: conlist(str, max_length=50) | None = None
    from_: str | None = Field(None, alias='from')
    sender: str | None = None
    subject: str | None = None
    tag: str | None = None
    reply_to: str | None = None
    plain_body: str | None = None
    html_body: str | None = None
    attachments: list[RequestAttachmentSchema] | None = None
    headers: dict[str, Any] | None = None
    bounce: bool | None = None


class RequestRawMessageSchema(BaseModel):
    mail_from: str
    rcpt_to: list[str]
    data: str
    bounce: bool | None = None



import logging
from typing import Any
from uuid import uuid4

from niquests import AsyncSession
from niquests.exceptions import ConnectTimeout

from .base import PostalPyAPIBase
from .exceptions import PostalPyConnectTimeoutError
from .schemas import (MessageExpansion,
                      RequestMessageDetailsSchema,
                      RequestMessageSchema,
                      RequestRawMessageSchema,
                      ResponseSchema)


class PostalPyAPI(PostalPyAPIBase):
    def __init__(self, base_url: str, api_key: str, timeout: int = 5, level: logging = logging.INFO):
        super().__init__(base_url=base_url, api_key=api_key, timeout=timeout, level=level)
        self._session = AsyncSession(base_url=self._base_url, timeout=self._timeout)
        self._session.headers = self._headers

    async def _send_request(self, url: str, json: dict[str, Any]) -> ResponseSchema:
        request_id = uuid4().hex
        self._logger.info('Request=%s url=%s json=%s', request_id, url, json)
        try:
            response = await self._session.post(url=url, json=json)
        except ConnectTimeout as e:
            self._logger.exception('Response=%s: %s', request_id, PostalPyConnectTimeoutError.__doc__)
            raise PostalPyConnectTimeoutError(e)
        return self._handle_response(response=response, request_id=request_id)

    async def get_message_details(self, data: RequestMessageDetailsSchema) -> ResponseSchema:
        """
        https://apiv1.postalserver.io/controllers/messages/message.html
        """
        url = '/api/v1/messages/message'
        expansions = [e for e in data.expansions or ()]
        if MessageExpansion.all in data.expansions:
            expansions = True
        json = {
            'id': data.id,
            '_expansions': expansions
        }
        return await self._send_request(url=url, json=json)

    async def get_message_deliveries(self, id: int) -> ResponseSchema:
        """
        https://apiv1.postalserver.io/controllers/messages/deliveries.html
        """
        url = '/api/v1/messages/deliveries'
        json = {'id': id}
        return await self._send_request(url=url, json=json)

    async def send_message(self, data: RequestMessageSchema) -> ResponseSchema:
        """
        https://apiv1.postalserver.io/controllers/send/message.html
        """
        url = '/api/v1/send/message'
        json = data.model_dump(exclude_none=True, by_alias=True)
        return await self._send_request(url=url, json=json)

    async def send_raw_message(self, data: RequestRawMessageSchema) -> ResponseSchema:
        """
        https://apiv1.postalserver.io/controllers/send/raw.html
        """
        url = '/api/v1/send/raw'
        json = data.model_dump(exclude_none=True)
        return await self._send_request(url=url, json=json)

    async def close(self):
        await self._session.close()
